/* GSI - general sound interface
 *
 * gsi_patchload_gus.c
 *
 * Copyright (C) W.H.Scholten 1998
 *
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <gsi/mytypes.h>
#include <gsi/config.h>
#include <gsi/debug.h>
#include <gsi/gsi_patchnames_gus.h>
#include "midi.h"

#ifdef __linux__
#  include <sys/soundcard.h>
#  include <sys/ultrasound.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#  include <machine/ultrasound.h>
#endif

#ifdef USE_OSS /* This code is currrently OSS only */

#ifdef USE_GUS_SYNTH


/*__________________DEFINITIES___________________*/

#ifndef GUS_PATCH_DIR
#  define GUS_PATCH_DIR "/dos/ultrasnd/midi/"
#endif


/*__________________EXTERN___________________*/
SEQ_USE_EXTBUF();


/*__________________FUNCTIE PROTOTYPES___________________*/
short lsb_int16_at(byte *p);
int   lsb_int32_at(byte *p);


/*__________________VARIABELEN___________________*/
char *gus_patch_dir=GUS_PATCH_DIR;

int  gus_memory;

static int  instrmap[256] = {-1, 0};
static int  initialized = 0;

/* This is the header as used in patch files */
struct pat_header {
	char            magic[12];
	char            version[10];
	char            description[60];
	unsigned char   instruments;
	char            voices;
	char            channels;
	unsigned short  nr_waveforms;
	unsigned short  master_volume;
	unsigned int    data_size;
};



/*__________________PROGRAMMA___________________*/
int gus_init(int seq_fd, int dev) {
	DEB1(fprt(stderr, "__GSI: gus_init, dev=%d\n", dev));

	gus_memory = dev;
	ioctl(seq_fd, SNDCTL_SYNTH_MEMAVL, &gus_memory);
	DEB1(fprt(stderr, "__GSI: GUS memory is %d\n", gus_memory));

	DEB1(fprintf(stderr, "dev=%d, card_info[dev].nr_voices=%d\n", dev, card_info[dev].nr_voices));
	if (ioctl (seq_fd, SNDCTL_SEQ_RESETSAMPLES, &dev) == -1) {
		fprintf(stderr, "GSI: Can't reset GM programs on device %d\n", dev);
		perror("GUS init");
		return -1;
	}


#if 1
	card_info[dev].nr_voices=24;
#else
	card_info[dev].nr_voices=16;
#endif

	GUS_NUMVOICES(dev, card_info[dev].nr_voices);
	if (voices[dev] != NULL) free(voices[dev]);
	voices[dev] = malloc(sizeof(struct voice_state) * card_info[dev].nr_voices);
	DEB1(fprintf(stderr, "dev=%d, card_info[dev].nr_voices=%d\n", dev, card_info[dev].nr_voices));
	initialized=1;
	return 0;
}


/*---------------------------------------------------------------------------*
 *	Load a patch, returns 0 if ok, -1 otherwise
 *	Currently does not do a check to see if loading failed when not enough
 *	memory is available on the sound card, due to the OSS macro's.
 *	(workaround: use write() or check to see if available memory changed))
 *---------------------------------------------------------------------------*/
int gus_load (int seq_fd, int type, int dev, int pgm) {
	int  i;
	int  patch_fd;
	char patch_filename[256];
	byte buf[256];
	int  offset;
	struct pat_header    header;
	struct sample_header sample;
	struct patch_info    *patch;

	DEB1(fprt(stderr, "__GSI: gus_load pgm=%d\n", pgm));

	if (!initialized)
		gus_init(seq_fd, dev);

	if (pgm<0 || pgm>255)
		return 0; /* ignore out of range values */

	if (instrmap[pgm] == pgm)
		return 0; /* already loaded */

	instrmap[pgm] = pgm;

	DEB1(fprt(stderr, "  patch name is %s\n", patch_names[pgm]));
	if (patch_names[pgm]==NULL || patch_names[pgm][0] == 0) /* Not defined */
		return 0;

	sprintf (patch_filename, "%s/%s.pat", gus_patch_dir, patch_names[pgm]);
	DEB1(fprintf(stderr, "  Loading instrument %d from patch file %s\n", pgm, patch_filename));

	if ((patch_fd = open (patch_filename, O_RDONLY, 0)) == -1)
	{
		perror (patch_filename);
		return -1;
	}

	if (read (patch_fd, buf, 0xef) != 0xef)
	{
		fprintf (stderr, "GSI: patch file %s is too short\n", patch_filename);
		return -1;
	}

	memcpy ((char *) &header, buf, sizeof (header));

	if (strncmp (header.magic, "GF1PATCH110", 12))
	{
		fprintf (stderr, "GSI: %s is not a patch file\n", patch_filename);
		return -1;
	}

	if (strncmp (header.version, "ID#000002", 10))
	{
		fprintf (stderr, "GSI: %s has incompatible patch file version\n", patch_filename);
		return -1;
	}

	header.nr_waveforms  = (uint16) lsb_int16_at(&buf[85]);
	header.master_volume = (uint16) lsb_int16_at(&buf[87]);

	offset = 0xef;

	for (i = 0; i < header.nr_waveforms; i++)
	{
		if (lseek (patch_fd, offset, 0) == -1)
		{
			perror (patch_filename);
			return -1;
		}

		if (read (patch_fd, &buf, sizeof (sample)) != sizeof (sample))
		{
			fprintf (stderr, "GSI: patch file %s is too short\n", patch_filename);
			return -1;
		}

		memcpy ((char *) &sample, buf, sizeof (sample));

		sample.fractions     = (char) buf[7];
		sample.len           = lsb_int32_at(&buf[8]);
		sample.loop_start    = lsb_int32_at(&buf[12]);
		sample.loop_end      = lsb_int32_at(&buf[16]);
		sample.base_freq     = (uint16) lsb_int16_at(&buf[20]);
		sample.low_note      = lsb_int32_at(&buf[22]);
		sample.high_note     = lsb_int32_at(&buf[26]);
		sample.base_note     = lsb_int32_at(&buf[30]);
		sample.detune        = lsb_int16_at(&buf[34]);
		sample.panning       = buf[36];

		memcpy (sample.envelope_rate, &buf[37], 6);
		memcpy (sample.envelope_offset, &buf[43], 6);

		sample.tremolo_sweep   = buf[49];
		sample.tremolo_rate    = buf[50];
		sample.tremolo_depth   = buf[51];

		sample.vibrato_sweep   = buf[52];
		sample.vibrato_rate    = buf[53];
		sample.vibrato_depth   = buf[54];
		sample.modes           = buf[55];
		sample.scale_frequency = lsb_int16_at(&buf[56]);
		sample.scale_factor    = (uint16) lsb_int16_at(&buf[58]);

		offset = offset + 96;
			/* Now allocate memory for the patch structure and sample data.
			 * The information in the temporary sample struct is then copied
			 * to the patch struct. */

#if 0
		gus_upload_sample(&sample);
	{}

void gus_upload_sample(&sample) {
}
#endif
		patch = (struct patch_info *) malloc (sizeof (*patch) + sample.len);
		if (patch == NULL) {
			fprintf(stderr, "GSI: Failed to allocate %d bytes of memory\n",
					sizeof (*patch) + sample.len);
			return -1;
		}

		patch->key = GUS_PATCH;
		patch->device_no = dev;
		patch->instr_no = pgm;

			/* Copy the temporary sample info to the patch struct */
		patch->mode = sample.modes | WAVE_TREMOLO | WAVE_VIBRATO | WAVE_SCALE;
		patch->len = sample.len;
		patch->loop_start = sample.loop_start;
		patch->loop_end = sample.loop_end;
		patch->base_note = sample.base_note;
		patch->high_note = sample.high_note;
		patch->low_note = sample.low_note;
		patch->base_freq = sample.base_freq;
		patch->detuning = sample.detune;
		patch->panning = (sample.panning - 7) * 16;

		memcpy (patch->env_rate, sample.envelope_rate, 6);
		memcpy (patch->env_offset, sample.envelope_offset, 6);

		patch->tremolo_sweep = sample.tremolo_sweep;
		patch->tremolo_rate = sample.tremolo_rate;
		patch->tremolo_depth = sample.tremolo_depth;

		patch->vibrato_sweep = sample.vibrato_sweep;
		patch->vibrato_rate = sample.vibrato_rate;
		patch->vibrato_depth = sample.vibrato_depth;

		patch->scale_frequency = sample.scale_frequency;
		patch->scale_factor = sample.scale_factor;


		patch->volume = header.master_volume;

			/* some sanity checking */
		if (lseek (patch_fd, offset, 0) == -1)
		{
			perror (patch_filename);
			return -1;
		}

		/* some more sanity checking */
		if (read (patch_fd, patch->data, sample.len) != sample.len)
		{
			fprintf (stderr, "GSI patch file %s is too short\n", patch_filename);
			return -1;
		}

		/* Send the patch to the sound card via OSS */
		SEQ_WRPATCH (patch, sizeof (*patch) + sample.len);
		free(patch);

		offset = offset + sample.len;
	}
	patchloaded[pgm]=1;

	close(patch_fd);
	gus_memory = dev;
	ioctl(seqfd, SNDCTL_SYNTH_MEMAVL, &gus_memory); /* update available memory */
	DEB1(fprt(stderr, "__GSI: gus memory after loading patch is %d\n", gus_memory)); 
	return 0;
}

#ifdef USE_OSS_38
int gus_unload (int seq_fd, int type, int dev, int pgm) {
	struct remove_sample sample;

	sample.devno=dev;
	sample.instrno=pgm &127;
	sample.bankno=(pgm>>7)&1;
	ioctl(seqfd, SNDCTL_SYNTH_REMOVESAMPLE, sample);

	return 0;
}
#endif

int lsb_int32_at(byte *p) {
	int  i;
	int  ret_val=0;

	for (i=0;i<4;i++)
		ret_val |= (p[i] << (i*8));

	return ret_val;
}

short lsb_int16_at(byte *p) {
	short ret_val=0;

	ret_val= p[0] | (((int) p[1]) <<8);

	return ret_val;
}
#endif /* USE_GUS_SYNTH */

#endif /* USE_OSS */

/*---------------------------------------------------------------------------*
 *	EOF gsi_patchload_gus.c
 *---------------------------------------------------------------------------*/
